/* graphtv.cpp : Chambolle-Darbon's implementation of D.~Hochbaum's algorithm
   for computing the prox of the total variation on a graph, based on 
   Y. Boykov and V. Kolmogorov's augmenting path algorithm for max flow.

   This must be compiled together with Boykov and Kolmogorov's implementation
   of the maxflow/mincut algorithm available at
   https://pub.ista.ac.at/~vnk/software.html (version 3.04)

   Please see README in that distribution and cite:
   Yuri Boykov and Vladimir Kolmogorov: An Experimental Comparison
         of Min-Cut/Max-Flow Algorithms for Energy Minimization in 
	 Computer Vision.
	 In IEEE Transactions on Pattern Analysis and Machine Intelligence,
	 September 2004.
   
   For the current program, please cite:

   D. S. Hochbaum: An efficient algorithm for image segmentation,
        Markov random fields and related problems. 
	J. ACM, 48(4):686--701, 2001.
	https://doi.org/10.1145/502090.502093
   Chambolle, A., Darbon, J.: On Total Variation Minimization and
        Surface Evolution Using Parametric Maximum Flows. 
	Int J Comput Vis 84, 288–307 (2009). 
	https://doi.org/10.1007/s11263-009-0238-9
   Chambolle, A., Darbon, J.: A parametric maximum flow approach 
        for discrete total variation regularization. 
	In: Image Processing and Analysis With Graphs: Theory and Practice.
	CRC Press. (2012)

   Former version (c) Antonin Chambolle Jérôme Darbon 2007.
   Current (slightly) improved version (c) Antonin Chambolle 2023.
   
   ***************************** important ********************************
   To compile: download maxflow-v3.04.src.zip from
   https://pub.ista.ac.at/~vnk/software.html
   ** at line 137 in graph.h ** add the line:
   void graphTV(tcaptype precision);
   ************************************************************************

   Then compile following the instructions in README.TXT
   graphTV.cpp provides the additional command G.graphTV(precision)
   which computes arg min ||u-g||^2/2 + E(u) where
   E(u) = sum_{arcs(i,j)} w_{i,j}(u_i-u_j)^+ (w = weight)
   g has to be provided in the tr_cap variable (use set_trcap to set)
   u is returned in the tr_cap variable (use get_trcap to read)
   The method finds "breakpoints" up to the precision "precision", which
   are the values taken by the solution u (precision = 0 is admissible).

   Usage: see the file tvh-exact.cpp
   
   This has *not* been tested with integer captypes and migth not work in
   such instances.

   The current file is a modification of maxflow.cpp in BK's distribution,
   where after each flow, the graph is modified to estimate new breakpoints.
*/

#include <stdio.h>
#include <stdlib.h>
#include "graph.h"

/*
	special constants for node->parent. Duplicated in graph.cpp, both should match!
*/
#define TERMINAL ( (arc *) 1 )		/* to terminal */
#define ORPHAN   ( (arc *) 2 )		/* orphan */


#define INFINITE_D ((int)(((unsigned)-1)/2))		/* infinite distance to the terminal */

/***********************************************************************/

template <typename captype, typename tcaptype, typename flowtype> 
void Graph<captype,tcaptype,flowtype>::graphTV(tcaptype precision)
{
	node *i, *j, *current_node = NULL;
	arc *a;
	nodeptr *np, *np_next;
	node_id I,J;
	#define LBSIZE 800
	int k,l,firstsink;
	int maxnumlabels = LBSIZE;
	// new variables for Hochbaum's algorithm
	node ***labellist, **nodelist;
	//node **lend;
	int *numl;
	tcaptype *values;
	tcaptype avm,avp; // should be tcaptype
	int lastlabel,nump,numm;
	//int bad=0, ite=0;
	bool *livelabel;

	if (!nodeptr_block)
	{
		nodeptr_block = new DBlock<nodeptr>(NODEPTR_BLOCK_SIZE, error_function);
	}

	// reuse_trees is considered FALSE here (see maxflow.cpp for details)
	changed_list = NULL;
	
	labellist = (node ***) malloc(maxnumlabels * sizeof(node **));
	numl = (int*) malloc(maxnumlabels * sizeof(int));
	values = (tcaptype *) malloc(maxnumlabels * sizeof(tcaptype));
	livelabel = (bool *) malloc(maxnumlabels * sizeof(bool));

	avm = 0;
	nodelist = (node**) malloc(node_num*sizeof(node *));
	for (k=0, i = nodes; k<node_num; k++,i++) {
	  nodelist[k]=i;
	  avm += i->tr_cap;
	}
	avm /= node_num;
	lastlabel = 0;
	values[lastlabel] = avm;
	labellist[lastlabel] = nodelist;
	numl[lastlabel] = node_num;
	livelabel[lastlabel]=true;
	
	for (i=nodes; i<node_last;i++) i->tr_cap -= avm;
	
	bool changed = true;
	while (changed) {
	  maxflow_init();
	  changed = false;

	  // main loop -> copy pasted from maxflow()
	  while ( 1 )
	    {
	      // test_consistency(current_node);

	      if ((i=current_node))
		{
		  i -> next = NULL; /* remove active flag */
		  if (!i->parent) i = NULL;
		}
	      if (!i)
		{
		  if (!(i = next_active())) break;
		}

	      /* growth */
	      if (!i->is_sink)
		{
		  /* grow source tree */
		  for (a=i->first; a; a=a->next)
		    if (a->r_cap)
		      {
			j = a -> head;
			if (!j->parent)
			  {
			    j -> is_sink = 0;
			    j -> parent = a -> sister;
			    j -> TS = i -> TS;
			    j -> DIST = i -> DIST + 1;
			    set_active(j);
			    add_to_changed_list(j);
			  }
			else if (j->is_sink) break;
			else if (j->TS <= i->TS &&
				 j->DIST > i->DIST)
			  {
			    /* heuristic - trying to make the distance from j to the source shorter */
			    j -> parent = a -> sister;
			    j -> TS = i -> TS;
			    j -> DIST = i -> DIST + 1;
			  }
		      }
		}
	      else
		{
		  /* grow sink tree */
		  for (a=i->first; a; a=a->next)
		    if (a->sister->r_cap)
		      {
			j = a -> head;
			if (!j->parent)
			  {
			    j -> is_sink = 1;
			    j -> parent = a -> sister;
			    j -> TS = i -> TS;
			    j -> DIST = i -> DIST + 1;
			    set_active(j);
			    add_to_changed_list(j);
			  }
			else if (!j->is_sink) { a = a -> sister; break; }
			else if (j->TS <= i->TS &&
				 j->DIST > i->DIST)
			  {
			    /* heuristic - trying to make the distance from j to the sink shorter */
			    j -> parent = a -> sister;
			    j -> TS = i -> TS;
			    j -> DIST = i -> DIST + 1;
			  }
		      }
		}

	      TIME ++;

	      if (a)
		{
		  i -> next = i; /* set active flag */
		  current_node = i;

		  /* augmentation */
		  augment(a);
		  /* augmentation end */

		  /* adoption */
		  while ((np=orphan_first))
		    {
		      np_next = np -> next;
		      np -> next = NULL;

		      while ((np=orphan_first))
			{
			  orphan_first = np -> next;
			  i = np -> ptr;
			  nodeptr_block -> Delete(np);
			  if (!orphan_first) orphan_last = NULL;
			  if (i->is_sink) process_sink_orphan(i);
			  else            process_source_orphan(i);
			}

			  orphan_first = np_next;
		    }
		  /* adoption end */
		}
	      else current_node = NULL;
	    }

	  // update graph and values
	  // ite++; for debugging: counts number of graphcuts
	  // first remove arcs from sink region to source region
	  for (a=arcs; a < arc_last; a++) {
	    get_arc_ends(a,I,J);
	    if (what_segment(I,SINK)==SINK && what_segment(J,SINK)==SOURCE)
	      set_rcap(a,0);
	  }
	  // then for each label check if region needs to be cut
	  for (l=lastlabel; l>=0; l--) if (livelabel[l]) {
	      //printf("label %d\n",l);
	      firstsink = 0;
	      avp = 0;
	      avm = 0;
	      nump = 0;
	      numm = 0;
	      for (k=0; k < numl[l]; k++) {
		// we will swap source and sink nodes in the label l,
		// putting (arbitrarily) source list first and then the sink
		// list which will be assigned a new label
		i = labellist[l][k];
		// see def of what_segment(node_id,SINK): test if source

		// the following commented line counts the number of pixels
		// which belong nor to sink or source ("undetermined").
		// Though this is rare (usually ~ 2% of processed pixels),
		// one could improve slightly the code by assigning to them
		// a dead label, the value value[l], and not processing
		// them any longer. They get back the correct value after
		// two more graphcuts.
		// if (!(i->parent)) bad++; // should be assigned value[l]
		
		if (i->parent && !(i->is_sink)) { // Source + undetermined
		  nump++;
		  avp += i->tr_cap;
		  if (k>firstsink) {
		    labellist[l][k] = labellist[l][firstsink];
		    labellist[l][firstsink] = i;
		  }
		  firstsink++;
		} else { 
		  numm++;
		  avm -= i-> tr_cap;
		}
	      }
	      if (nump) avp /= nump;
	      if (numm) avm /= numm;

	      if (nump == 0 || numm == 0 || (avp <= precision && avm <= precision))
		{
		  livelabel[l] = false; // cut off this label which is finished
		  // set to 0 trcaps [probably useless and adds O(N) cost]
		  for (k=0; k < numl[l]; k++) labellist[l][k]->tr_cap = 0;
		}
	      else { // increase the number of labels
		changed = true;
		lastlabel++;
		if (lastlabel >= maxnumlabels) {
		  // realloc memory : labellist, livelabel, values, numl
		  maxnumlabels += LBSIZE;
		  labellist = (node ***) realloc(labellist,maxnumlabels * sizeof(node **));
		  numl = (int *) realloc(numl,maxnumlabels * sizeof(int));
		  values = (tcaptype *) realloc(values,maxnumlabels * sizeof(tcaptype));
		  livelabel = (bool *) realloc(livelabel,maxnumlabels * sizeof(bool));
		}
		// create new label with sink nodes
		labellist[lastlabel] = labellist[l]+firstsink;
		numl[l] = nump; // = firstsink
		numl[lastlabel] = numm;
		livelabel[lastlabel] = true;
		values[lastlabel] = values[l] - avm;
		values[l] += avp;
		// adjust tr_caps for the two regions
		for (k=0; k<numl[l]; k++) {
		  labellist[l][k]->tr_cap -= avp;
		}
		for (k=0; k<numl[lastlabel]; k++) {
		  labellist[lastlabel][k]->tr_cap += avm;
		}
	      }
	    }
	}
	
	{
		delete nodeptr_block; 
		nodeptr_block = NULL; 
	}

	// set values back in the tr_cap variable
	for (l=0; l <= lastlabel; l++) {
	  avm = values[l];
	  for (k=0; k<numl[l]; k++) {
	    labellist[l][k]->tr_cap = avm;
	  }
	}

	free(labellist);
	free(numl);
	free(values);
	free(livelabel);
	//printf("it: %d, bad labels: %d (%lf)\n",ite,bad,((double)bad)/node_num);
}

/***********************************************************************/
#include "instances.inc"
